import numpy as np
import pyarrow as pa
from wombat_db.ops.helpers import columns_to_array, groupify_array

# Filter functionality
def arr_op_to_idxs(arr, op, value):
    # Cast value to type arr
    try:
        value = np.array(value, dtype=arr.dtype)
    except:
        raise Exception("Cannot downcast {} to data type {}".format(value, arr.dtype))

    if op in ['=', '==']:
        return np.where(arr == value)
    elif op == '!=':
        return np.where(arr != value)
    elif op == '<':
        return np.where(arr < value)
    elif op == '>':
        return np.where(arr > value)
    elif op == '<=':
        return np.where(arr <= value)
    elif op == '>=':
        return np.where(arr >= value)
    elif op == 'in':
        mask = np.isin(arr, value)
        return np.arange(len(arr))[mask]
    elif op == 'not in':
        mask = np.invert(np.isin(arr, value))
        return np.arange(len(arr))[mask]
    else:
        raise Exception("Operand {} is not implemented!".format(op))

def filters(table, filters):
    filters = ([filters] if isinstance(filters, tuple) else filters)
    # Filter is a list of (col, op, value) tuples
    idxs = np.arange(table.num_rows)
    for (col, op, value) in filters: #= or ==, !=, <, >, <=, >=, in and not in
        arr = table.column(col).to_numpy()
        f_idxs = arr_op_to_idxs(arr[idxs], op, value)
        idxs = idxs[f_idxs]
    return table.take(idxs)

# Drop duplicates
def drop_duplicates(table, on=[], keep='first'):
    # Gather columns to arr
    arr = columns_to_array(table, (on if on else table.column_names))

    # Groupify
    dic, counts, sort_idxs, bgn_idxs = groupify_array(arr)

    # Gather idxs
    if keep == 'last':
        idxs = (np.array(bgn_idxs) - 1)[1:].tolist() + [len(sort_idxs) - 1]
    elif keep == 'first':
        idxs = bgn_idxs
    elif keep == 'drop':
        idxs = [i for i, c in zip(bgn_idxs, counts) if c == 1]
    return table.take(sort_idxs[idxs])

# Show for easier printing
def head(table, n=5, max_width=100):
    if table.num_rows == 0:
        print("No data in table")
        return
    
    # Extract head data
    t = table.slice(length=n)
    head = {k: list(map(str, v)) for k, v in t.to_pydict().items()}

    # Calculate width
    col_width = list(map(len, head.keys()))
    data_width = [max(map(len, h)) for h in head.values()]

    # Print data
    data = [list(head.keys())] + [[head[c][i] for c in head.keys()] for i in range(t.num_rows)]
    for i in range(len(data)):
        adjust = [w.ljust(max(cw, dw) + 2) for w, cw, dw in zip(data[i], col_width, data_width)]
        print(('Row  ' if i == 0 else str(i-1).ljust(5)) + "".join(adjust)[:max_width])
    print('\n')


import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

def groupify_array(arr):
    # Input: Pyarrow/Numpy array
    # Output:
    #   - 1. Unique values
    #   - 2. Sort index
    #   - 3. Count per unique
    #   - 4. Begin index per unique
    dic, counts = np.unique(arr, return_counts=True)
    sort_idx = np.argsort(arr)
    return dic, counts, sort_idx, [0] + np.cumsum(counts)[:-1].tolist()

def combine_column(table, name):
    return table.column(name).combine_chunks()

def _dictionary_and_indices(column):
    assert isinstance(column, pa.ChunkedArray)

    if not isinstance(column.type, pa.DictionaryType):
        column = pc.dictionary_encode(column, null_encoding_behavior='encode')

    dictionary = column.chunk(0).dictionary
    indices = pa.chunked_array([c.indices for c in column.chunks])

    if indices.null_count != 0:
        # We need nulls to be in the dictionary so that indices can be
        # meaningfully multiplied, so we must round trip through decoded
        column = pc.take(dictionary, indices)
        return _dictionary_and_indices(column)

    return dictionary, indices

def columns_to_array(table, columns):
    columns = ([columns] if isinstance(columns, str) else list(set(columns)))
    combined_indices = None
    for c in columns:
        dictionary, indices = _dictionary_and_indices(table.column(c))
        if combined_indices is None:
            combined_indices = indices
        else:
            combined_indices = pc.add(
                pc.multiply(combined_indices, len(dictionary)),
                indices
            )
    return combined_indices.to_numpy()

def tables_to_arrays(table1, table2, columns):
    columns = ([columns] if isinstance(columns, str) else list(set(columns)))
    combined_indices = None
    len1, len2 = table1.num_rows, table2.num_rows 
    for c in columns:
        arr1, arr2 = combine_column(table1, c), combine_column(table2, c)
        dictionary, indices = _dictionary_and_indices(pa.chunked_array(pa.concat_arrays([arr1, arr2.cast(arr1.type)])))
        if combined_indices is None:
            combined_indices = indices
        else:
            combined_indices = pc.add(
                pc.multiply(combined_indices, len(dictionary)),
                indices
            )
    return combined_indices[:len1].to_numpy(), combined_indices[len1:].to_numpy()

# Old helpers

# Splitting tables by columns
def split_array(arr):
    arr = arr.dictionary_encode()
    ind, dic = arr.indices.to_numpy(zero_copy_only=False), arr.dictionary.to_numpy(zero_copy_only=False)

    if len(dic) < 1000:
        # This method is much faster for small amount of categories, but slower for large ones
        return {v: (ind == i).nonzero()[0] for i, v in enumerate(dic)}
    else:
        idxs = [[] for _ in dic]
        [idxs[v].append(i) for i, v in enumerate(ind)]
        return dict(zip(dic, idxs))

def split(table, columns, group=(), idx=None):
    # idx keeps track of the orginal table index, getting split recurrently
    if not isinstance(idx, np.ndarray):
        idx = np.arange(table.num_rows)
    val_idxs = split_array(combine_column(table, columns[0]))
    if columns[1:]:
        return [s for v, i in val_idxs.items() for s in split(table, columns[1:], group + (v,), idx[i])]
    else:
        return [(group + (v,), i) for v, i in val_idxs.items()]


from wombat_db.engine.nodes import *
from wombat_db.engine.sql import parse_sql
from wombat_db.engine.column import ColumnNode

# Computation plan (of multiple nodes)
class ExecutionPlan():
    def __init__(self, node):
        self.database, self.last, self.cache_obj = node.database, node, node.cache_obj

    def __getitem__(self, key):
        if isinstance(key, ColumnNode):
            if key.boolean:
                self.last = BooleanMaskNode(self.last, key, self.cache_obj)
            else:
                raise Exception('Column must be boolean to be used in selection')
        elif key in self.last.columns:
            return ColumnNode(key, required=[key])
        else:
            raise Exception("{} not a valid column reference")

    def __setitem__(self, key, value):
        if isinstance(value, ColumnNode):
            self.last = CalculationNode(self.last, key, value, cache_obj=self.cache_obj)
        else:
            raise Exception("Value must be a column node reference")

    def collect(self, verbose=False):
        if verbose:
            print("Columns:", ", ".join(self.last.columns_forward))
        self.last.backward(columns_backward=self.last.columns_forward, filters_backward=self.last.filters_forward)
        return self.last.get(verbose)

    # Numerical operations
    def filter(self, filters):
        self.last = FilterNode(self.last, filters, cache_obj=self.cache_obj)
        return self

    def join(self, right, on):
        if isinstance(right, str):
            plan = self.database.select(right)
            self.last = JoinNode(self.last, plan.last, on, cache_obj=self.cache_obj)
        else:
            self.last = JoinNode(self.last, right.last, on, cache_obj=self.cache_obj)
        return self

    def aggregate(self, by, methods):
        self.last = AggregateNode(self.last, by, methods, cache_obj=self.cache_obj)
        return self

    def rename(self, mapping):
        self.last = SelectionNode(self.last, list(mapping.keys()), aliases=list(mapping.values()), cache_obj=self.cache_obj)
        return self

    def select(self, columns=[]):
        if columns:
            self.last = SelectionNode(self.last, columns, cache_obj=self.cache_obj)
        else:
            self.last = SelectionNode(self.last, self.last.columns, cache_obj=self.cache_obj)
        return self

    def drop(self, columns=[]):
        self.last = DropNode(self.last, columns, cache_obj=self.cache_obj)
        return self

    def orderby(self, key, ascending=True):
        self.last = OrderNode(self.last, key, ascending, cache_obj=self.cache_obj)
        return self

    def fillna(self, columns, value):
        self.last = FillNanNode(self.last, columns, value, cache_obj=self.cache_obj)
        return self

    def cast(self, dtypes):
        self.last = CastNode(self.last, dtypes, cache_obj=self.cache_obj)
        return self

    def udf(self, name, arguments):
        return ColumnNode.udf(name=name, function=self.database.udfs[name], arguments=arguments)

    # Properties
    def columns(self):
        return self.last.columns

    # Other utilities
    def plot(self, name):
        from graphviz import Digraph
        dot = Digraph()
        nodes, count = [self.last], 1
        dot.node(str(count), label=self.last.graph_info(), shape='box')
        while nodes:
            if hasattr(nodes[0], 'parent'):
                count += 1
                dot.node(str(count), label=nodes[0].parent.graph_info(), shape='box')
                dot.edge(str(count), str(count - 1))
                nodes[0] = nodes[0].parent
            elif hasattr(nodes[0], 'left'):
                # Right first
                count += 1
                dot.node(str(count), label=nodes[0].right.graph_info(), shape='box')
                dot.edge(str(count), str(count - 1))
                nodes.append(nodes[0].right)

                # Left     
                count += 1   
                dot.node(str(count), label=nodes[0].left.graph_info(), shape='box')
                dot.edge(str(count), str(count - 2))
                nodes[0] = nodes[0].left
            else:
                nodes.pop(0)
        dot.render('plan/{}'.format(name), view=True)
        return

class Cache():
    def __init__(self, max_memory=1e9):
        self.tables, self.importance, self.memory, self.max_memory = {}, {}, 0, max_memory

    def put(self, key, table, weight=1.0):
        self.importance[key] = self.importance.get(key, 0.0) + weight
        if key not in self.tables.keys():
            b = table.nbytes
            while True:
                if self.memory + b < self.max_memory:
                    self.tables[key] = table
                    self.memory += b
                    return
                
                keys = list(self.tables.keys())
                importances = [self.importance[k] for k in keys]
                if not importances:
                    return
                
                if self.importance[key] > min(importances):
                    min_key = keys[importances.index(min(importances))]
                    self.memory -= self.tables[min_key]
                    del self.tables[min_key]
                else:
                    return

    def keys(self):
        return self.tables.keys()

    def __getitem__(self, key):
        return self.tables[key]

class Engine():
    def __init__(self, cache_memory=0):
        self.cache, self.tables, self.datasets, self.udfs = (cache_memory > 0), {}, {}, {}
        self.cache_obj = (Cache(max_memory=cache_memory) if self.cache else None)

    def register_table(self, name, table):
        self.tables[name] = table

    def register_dataset(self, name, dataset):
        self.datasets[name] = dataset

    def register_udf(self, name, function):
        self.udfs[name] = function

    def __getitem__(self, key):
        return self.select(key)

    def select(self, name):
        # Return a plan from a source node
        if name in self.tables.keys():
            return ExecutionPlan(TableNode(name, self, cache_obj=self.cache_obj))
        elif name in self.datasets.keys():
            return ExecutionPlan(DatasetNode(name, self, cache_obj=self.cache_obj))
        else:
            raise Exception("{} not in registered tables or datasets".format(name))

    def sql(self, sql):
        # Parse subqueries
        return parse_sql(self, sql)



import numpy as np
import pyarrow as pa
from wombat_db.ops.helpers import combine_column, columns_to_array, groupify_array

# Grouping / groupby methods
agg_methods = {
    'sum': np.sum,
    'max': np.max,
    'min': np.min,
    'mean': np.mean,
    'median': np.median,
    'count': np.size,
    'distinct_count': lambda a: np.unique(a).size,
    'distinct_sets': lambda a: len(set(a)),
    'prod': np.prod,
    'std': np.std,
    'var': np.var,
    'first': lambda a: a[0],
    'last': lambda a: a[-1],
}
def add_agg_method(self, name, method):
    def f(agg_columns=[]):
        methods = {col: method for col in (agg_columns if agg_columns else self.table.column_names) if col not in self.columns}
        return self.aggregate(methods=methods)
    setattr(self, name, f)

class Grouping():
    def __init__(self, table, columns):
        self.table = table
        self.columns = list(set(columns))

        # Initialize array + groupify
        self.arr = columns_to_array(table, columns)
        self.dic, self.counts, self.sort_idxs, self.bgn_idxs = groupify_array(self.arr)
        self.set_methods()

    def __iter__(self):
        for i in range(len(self.dic)):
            idxs = self.sort_idxs[self.bgn_idxs[i] : self.bgn_idxs[i] + self.counts[i]]
            yield {k: v[0] for k, v in self.table.select(self.columns).take([self.sort_idxs[self.bgn_idxs[i]]]).to_pydict().items()}, self.table.take(idxs)

    # Aggregation methods
    def set_methods(self):
        for k, m in agg_methods.items():
            add_agg_method(self, k, m)

    def aggregate(self, methods):
        # Create index columns
        table = self.table.select(self.columns).take(self.sort_idxs[self.bgn_idxs])
        self.refs = list(set(c for c, _ in methods.values()))
        data = {k: self.table.column(k).to_numpy() for k in self.refs}
        for col, (ref, f) in methods.items():
            vf = np.vectorize(f, otypes=[object])
            agg_arr = vf(np.split(data[ref][self.sort_idxs], self.bgn_idxs[1:]))
            table = table.append_column(col, pa.array(agg_arr))
        return table

    def agg(self, methods):
        methods = {col: ((m[0], agg_methods[m[1]]) if isinstance(m, tuple) else (col, agg_methods[m])) for col, m in methods.items()}
        return self.aggregate(methods=methods)

def groupby(table, by):
    return Grouping(table, by)


import re

def analyse_select_part(select):
	aggregate = '(' in select
	pcs = [s for s in select.split(' ') if s != '']
	column = (pcs[0] if not aggregate else pcs[0][pcs[0].find('(') + 1 : pcs[0].find(')')])
	method = (pcs[0][:pcs[0].find('(')] if aggregate else None)
	alias = (pcs[pcs.index('as') + 1] if 'as' in pcs else None)
	return column, alias, method


def parse_subquery(db, sql):
	# 1. FROM -> 2. JOINs -> 3. WHERE -> 4. CALCULATE -> 5. AGGREGATE -> 6. CALCULATE -> 7. HAVING -> 8. SELECT -> 9. ORDER BY -> 10. LIMIT
	sqll = re.sub(r"[\n\t]*", "", sql.lower()) # re.sub(' +', ' ', 
	print(sqll)

	# 0. Split query by keys
	keys = ['select', 'from', 'where', 'group by', 'having', 'order by', 'limit']
	idxs = [sqll.find(k) for k in keys]

	# Gather texts
	keys, idxs = zip(*[(k, i) for k, i in zip(keys, idxs) if i != -1])
	parts = {k:sqll[i + len(k) + 1 : j] for k, i, j in zip(keys, idxs, idxs[1:] + (None,))}

	# 1. Extract JOINs out FROM
	from_sql = parts['from']
	join_keys = ['join', 'inner join']
	join_idxs = [[m.start() for m in re.finditer(k, from_sql.lower())] for k in join_keys]

	# 1a. FROM
	from_end = min([i for fi in join_idxs for i in fi] + [len(from_sql)])
	if re.sub(' ', '', from_sql)[0] == '(':
		plan = parse_subquery(from_sql[1:from_end - 1])
	else:
		from_table = re.sub(' ', '', from_sql[0:from_end])
		plan = db.select(from_table)

	# 2. JOINs
	join_idxs_flat = sorted([i for join_idx in join_idxs for i in join_idx])
	join_parts = {i:from_sql[i:j] for i, j in zip(join_idxs_flat, join_idxs_flat[1:] + [None])}

	for k, join_idx in zip(join_keys, join_idxs):
		for i in join_idx:
			# Analyse the join sql
			join_sql = re.sub(' +', ' ', join_parts[i][len(k) + 1:])
			join_table = [s for s in join_sql.split(' ') if s != ''][0]
			if join_table[0] == '(':
				join_plan = parse_subquery(join_sql)
			else:
				join_plan = db.select(join_table)
			
			# Analyse whether USING / ON is used
			if 'on' in join_sql.lower():
				raise Exception("Not implemented: ON")
			elif 'using' in join_sql.lower():
				on = join_sql[join_sql.find('(') + 1: join_sql.find(')')].replace(' ', '').split(',')

			# Perform JOIN in the plan
			plan = plan.join(join_plan, on=on)
	
	# 3. WHERE
	if 'where' in parts.keys():
		where_sql = parts['where']
		if ' or ' in where_sql:
			raise Exception('Not implemented')
		else:
			where_clauses = where_sql.replace("'", "").split(' and ')
			filt = [tuple(c.split(' ')[:3]) for c in where_clauses]
			print(filt)
			plan = plan.filter(filt)

	# Gather selections
	selections = [analyse_select_part(s) for s in parts['select'].split(',')]
	print(selections)

	# 5. GROUP BY AGGREGATE
	if 'group by' in parts.keys():
		by = parts['group by'].replace(' ', '').split(',')
		methods = {col:method for col, alias, method in selections if method}
		plan = plan.aggregate(by=by, methods=methods)

	# 9. ORDER BY
	if 'order by' in parts.keys():
		orderby = [[v for v in p.split(' ') if v != ''] for p in parts['order by'].split(',')]
		plan = plan.orderby(orderby[0][0], ascending=(orderby[0][1] == 'asc' if len(orderby[0]) > 1 else True))

	return plan

def match_parenthesis(text):
	start, d = [], {}
	for i, c in enumerate(text):
		if c == '(':
			istart.append(i)
		if c == ')':
			try:
				d[istart.pop()] = i
			except IndexError:
				print('Too many closing parentheses')
	return d

def parse_sql(db, sql):
	return parse_subquery(db, sql)


from setuptools import find_packages, setup
from setuptools import Extension
import numpy as np
from Cython.Build import cythonize

__version__ = "0.0.17"

extensions = [
    Extension(
        name="cjoin", 
        sources=["wombat_db/ops/cjoin.c"], 
        include_dirs=[np.get_include()]
    )
]

with open('README.md') as readme_file:
    README = readme_file.read()

setup(
    name='wombat_db',
    version=__version__,
    description='Useful data crunching tools for pyarrow',
    long_description_content_type="text/markdown",
    long_description=README,
    license='APACHE',
    packages=find_packages(),
    author='Tom Scheffers',
    author_email='tom@youngbulls.nl ',
    keywords=['arrow', 'pyarrow', 'data', 'sql', 'dataframe'],
    url='https://github.com/TomScheffers/wombat',
    download_url='https://pypi.org/project/wombat-db/',

    ext_modules=cythonize(extensions),
    install_requires=[
        'numpy>=1.19.2',
        'pyarrow>=5.0.0'
    ],
)


import numpy as np
import pyarrow as pa
import json, time

def json_to_struct(arr):
    return pa.array(np.vectorize(json.loads)(arr))

def text_extract(text, key):
    value = '"' + key + '":'
    start = text.find(value)
    if start > -1:
        end = text[start:].find((',' if ',' in text[start:] else '}'))
        return eval(text[start + len(value): start + end])
    else:
        return None

def json_extract(arr, key):
    f = np.vectorize(lambda t: text_extract(t, key))
    return pa.array(f(arr.to_numpy()))


import pyarrow as pa
import pyarrow.parquet as pq
import numpy as np
from wombat_db.ops import join, groupby, filters
from wombat_db.engine.column import ColumnNode
import hashlib, json, time

# Computation nodes
class BaseNode():
    def check(self, needed, reference):
        missing = [c for c in needed if c not in reference]
        if missing:
            raise Exception("Required columns ({}) not in available columns ({})".format(", ".join(missing), ", ".join(reference)))

    def columns_bw(self, columns_backward):
        self.columns_backward = [c for c in sorted(list(set(self.columns_forward + columns_backward))) if c in self.columns_source]
        return self.columns_backward

    def properties(self):
        fields = ['table', 'on', 'filters', 'by', 'methods', 'key', 'ascending', 'calculation', 'columns_backward']
        obj = {k: v for k,v in self.__dict__.items() if k in fields}
        return {**{'name': self.__class__.__name__}, **obj}

    def graph_info(self):
        obj = self.properties()
        return "\\l".join([k + ': ' + str(v).replace("'", "") for k,v in obj.items() if k not in ['columns_backward']])

    def hash(self, h=None):
        # We want to add: on, by, columns_backward, table_url
        if not h:
            h = hashlib.sha256()
        h.update(json.dumps(self.properties(), sort_keys=True).encode())
        self.hash_key = h.hexdigest()
        return h

    def backward(self, columns_backward=[], filters_backward=[]):
        self.columns_bw(columns_backward)
        hp = self.parent.backward(columns_backward=self.columns_backward, filters_backward=filters_backward)
        return self.hash(h=hp)

    def get(self, verbose):
        self.time = time.time()
        if self.cache and self.hash_key in self.cache_obj.keys():
            t = self.cache_obj[self.hash_key]
            if verbose:
                print("Node: {} Rows: {} Cumulative Time: {:2f} (cached)".format(self.__class__.__name__.ljust(16), str(t.num_rows).ljust(9), time.time() - self.time))
        else:
            t = self.fetch(verbose)
            if verbose:
                print("Node: {} Rows: {} Cumulative Time: {:2f}".format(self.__class__.__name__.ljust(16), str(t.num_rows).ljust(9), time.time() - self.time))
            if self.cache:
                self.cache_obj.put(self.hash_key, t, weight=time.time() - self.time)
        return t

# Sources
class TableNode(BaseNode):
    def __init__(self, table, database, cache_obj=None):
        self.table, self.database, self.cache_obj = table, database, cache_obj
        self.cache = (cache_obj != None)
        self.t = self.database.tables[table]
        self.columns = self.t.column_names
        self.columns += list(set([c.split('.')[0] for c in self.columns if '.' in c]))

        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = self.columns, [], []

    def backward(self, columns_backward=[], filters_backward=[]):
        self.columns_bw(columns_backward)
        self.filters = filters_backward
        return self.hash()

    def fetch(self, verbose):
        t = self.t.select(self.columns_backward)
        tf = (filters(t, self.filters) if self.filters else t)
        return tf

def part_check(part, op, value):
    # Try to cast partition to value
    try:
        part = type(value)(part)
    except:
        raise Exception("Cannot downcast {} to data type {}".format(part, type(value)))

    if op in ['=', '==']:
        return part == value
    elif op == '!=':
        return part != value
    elif op == '<':
        return part < value
    elif op == '>':
        return part > value
    elif op == '<=':
        return part <= value
    elif op == '>=':
        return part >= value
    elif op == 'in':
        return part in value
    elif op == 'not in':
        return part not in value
    else:
        raise Exception("Operand {} is not implemented!".format(op))

class DatasetNode(BaseNode):
    def __init__(self, table, database, cache_obj=None):
        self.table, self.database, self.cache_obj = table, database, cache_obj
        self.cache = (cache_obj != None)
        self.dataset = database.datasets[table]

        self.partition_keys = [p.name for p in self.dataset.partitions]
        self.partition_values = [{pk[0]: dp.keys[pk[1]] for pk, dp in zip(p.partition_keys, self.dataset.partitions)} for p in self.dataset.pieces]
        self.meta = self.dataset.pieces[0].get_metadata()
        self.columns = self.partition_keys + [c['path_in_schema'] for c in self.meta.row_group(0).to_dict()['columns']]
        self.columns += list(set([c.split('.')[0] for c in self.columns if '.' in c]))

        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = self.columns, [], []

    def backward(self, columns_backward=[], filters_backward=[]):
        self.columns_bw(columns_backward)
        self.filters = filters_backward
        self.part_filters = list(filter(lambda f: f[0] in self.partition_keys, self.filters))
        self.value_filters = list(filter(lambda f: f[0] not in self.partition_keys, self.filters))
        return self.hash()

    def partition_check(self, partition_value, filters):
        for key, op, value in filters:
            if not part_check(partition_value[key], op, value):
                return False
        return True

    def fetch(self, verbose):
        ts = []
        for i, p in enumerate(self.dataset.pieces):
            if self.partition_check(self.partition_values[i], self.part_filters):
                t = p.read(
                    columns=[c for c in self.columns_backward if c not in self.partition_keys], 
                    partitions=self.dataset.partitions
                )
                ts.append(t)
        table = pa.concat_tables(ts)
        return (filters(table, self.value_filters) if self.value_filters else table)

    def fetch_v2(self, verbose):
        ts = []
        columns = [c for c in self.columns_backward if c not in self.partition_keys]
        read_filters, final_filters = [v for v in self.value_filters if '.' not in v[0]], [v for v in self.value_filters if '.' in v[0]]        
        for i, p in enumerate(self.dataset.pieces):
            if self.partition_check(self.partition_values[i], self.part_filters):
                # Read filtered parquet tables
                t = pq.read_table(
                    source=p.path,
                    columns=columns, 
                    use_legacy_dataset=False,
                    filters=read_filters
                )

                # Add Partitions & append
                for k, v in  self.partition_values[i].items():
                    t = t.append_column(k, pa.array([v for i in range(t.num_rows)]))
                ts.append(t)  
        table = pa.concat_tables(ts)
        return (filters(table, final_filters) if final_filters else table)

# Operations
def column_min_max(arr):
    if hasattr(arr, 'dictionary'):
        mmx = pa.compute.min_max(arr.indices)
        return (arr.dictionary[mmx['min'].as_py()].as_py(), arr.dictionary[mmx['max'].as_py()].as_py())
    else:
        mmx = pa.compute.min_max(arr)
        return (mmx['min'].as_py(), mmx['max'].as_py())

class JoinNode(BaseNode):
    def __init__(self, left, right, on, cache_obj=None):
        self.left, self.right, self.on, self.cache_obj = left, right, (on if isinstance(on, list) else [on]), cache_obj
        self.cache = (cache_obj != None)

        # Check columns
        self.check(needed=self.on, reference=left.columns)
        self.check(needed=self.on, reference=right.columns)
        self.columns = list(set(left.columns + right.columns))

        # Forward propagation of nodes
        self.columns_source, self.columns_forward = list(set(left.columns_source + right.columns_source)), list(set(left.columns_forward + right.columns_forward + on))
        self.filters_forward = left.filters_forward + right.filters_forward #[f for f in left.filters_forward + right.filters_forward if f[0] in self.on]

    def backward(self, columns_backward=[], filters_backward=[]):
        self.columns_bw(columns_backward)
        self.filters = filters_backward
        filters_l, filters_r = [f for f in self.filters if f[0] in self.left.columns_source], [f for f in self.filters if f[0] in self.right.columns_source]
        columns_l, columns_r = [c for c in self.columns_backward if c in self.left.columns_source], [c for c in self.columns_backward if c in self.right.columns_source]
        hl = self.left.backward(columns_backward=columns_l, filters_backward=filters_l)
        hr = self.right.backward(columns_backward=columns_r, filters_backward=filters_r)
        hl.update(hr.digest())
        return self.hash(h=hl)

    def fetch(self, verbose):
        tl = self.left.get(verbose)
        tr = self.right.get(verbose)
        return join(left=tl, right=tr, on=self.on)

class FilterNode(BaseNode):
    def __init__(self, parent, filters, cache_obj=None):
        self.parent, self.filters, self.cache_obj = parent, ([filters] if isinstance(filters, tuple) else filters), cache_obj
        self.cache = (cache_obj != None)

        # Check if columns are available
        self.check(needed=[f[0] for f in self.filters], reference=parent.columns)
        self.columns = parent.columns

        # Forward propagation of nodes
        self.columns_source, self.columns_forward = parent.columns_source, list(set(parent.columns_forward + [f[0] for f in self.filters if f[0] in parent.columns_source]))
        self.filters_forward = parent.filters_forward + self.filters

    def backward(self, columns_backward=[], filters_backward=[]):
        self.columns_bw(columns_backward)
        hp = self.parent.backward(columns_backward=self.columns_backward, filters_backward=filters_backward)
        self.hash_key = hp.hexdigest() # Filter node does not change anything, so can just pass its parents hash
        return hp

    def fetch(self, verbose):
        return self.parent.get(verbose)

class AggregateNode(BaseNode):
    def __init__(self, parent, by, methods, cache_obj=None):
        self.parent, self.by, self.methods, self.filters, self.cache_obj = parent, (by if isinstance(by, list) else [by]), methods, [], cache_obj
        self.cache = (cache_obj != None)

        # Check if columns are available
        refs = [(m[0] if isinstance(m, tuple) else k) for k, m in methods.items()]
        self.check(needed=self.by + refs, reference=parent.columns)
        self.columns = self.by + list(methods.keys())

        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = parent.columns_source, list(set(parent.columns_forward + [c for c in self.by + refs if c in parent.columns_source])), parent.filters_forward

    def backward(self, columns_backward=[], filters_backward=[]):
        self.columns_bw(columns_backward)
        self.filters = [f for f in filters_backward if f[0] in self.methods.keys()] # Intercept filters which are aggregate values
        hp = self.parent.backward(columns_backward=self.columns_backward, filters_backward=[f for f in filters_backward if f not in self.filters])
        return self.hash(h=hp)

    def fetch(self, verbose):
        tp = self.parent.get(verbose)
        t = groupby(tp, self.by).agg(self.methods)
        return (filters(t, self.filters) if self.filters else t)

class OrderNode(BaseNode):
    def __init__(self, parent, key, ascending, cache_obj=None):
        self.parent, self.key, self.ascending, self.cache_obj = parent, key, ascending, cache_obj
        self.cache = (cache_obj != None)

        # Check if columns are available
        self.check(needed=[key], reference=parent.columns)
        self.columns = parent.columns 
    
        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = parent.columns_source, list(set(parent.columns_forward + ([self.key] if self.key in parent.columns_source else []))), parent.filters_forward

    def fetch(self, verbose):
        tp = self.parent.get(verbose)
        idxs = pa.compute.sort_indices(tp.column(self.key)).to_numpy()
        return (tp.take(idxs) if self.ascending else tp.take(idxs[::-1]))

class SelectionNode(BaseNode):
    def __init__(self, parent, columns=[], aliases=[], cache_obj=None):
        self.parent, self.columns, self.aliases, self.cache_obj = parent, columns, aliases, cache_obj
        if not columns:
            self.columns = [c for c in self.parent.columns if '.' not in c]
        self.cache = (cache_obj != None)
        self.mapping = (dict(zip(columns, aliases)) if aliases else {})

        # Check if columns are available
        self.check(needed=columns, reference=parent.columns)
        if aliases: 
            self.columns = [self.mapping.get(c, c) for c in parent.columns]

        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = parent.columns_source, list(set(parent.columns_forward + [c for c in columns if c in parent.columns_source])), parent.filters_forward
        
    def fetch(self, verbose):
        tp = self.parent.get(verbose)
        if self.aliases:
            return tp.rename_columns([self.mapping.get(col, col) for col in tp.column_names])
        else:
            return tp.select(self.columns)

class CalculationNode(BaseNode):
    def __init__(self, parent, key, column, cache_obj=None):
        self.parent, self.key, self.calculation, self.column, self.cache_obj = parent, key, column.key, column, cache_obj
        self.cache = (cache_obj != None)

        # Check if columns are available
        self.check(needed=column.required, reference=parent.columns)
        self.columns = parent.columns + [key]
    
        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = parent.columns_source, list(set(parent.columns_forward + [c for c in column.required if c in parent.columns_source])), parent.filters_forward
    
    def backward(self, columns_backward=[], filters_backward=[]):
        self.columns_bw(columns_backward)
        self.filters = [f for f in filters_backward if f[0] == self.key] # Intercept filters which are calculated values
        hp = self.parent.backward(columns_backward=self.columns_backward, filters_backward=[f for f in filters_backward if f not in self.filters])
        return self.hash(h=hp)

    def fetch(self, verbose):
        tp = self.parent.get(verbose)
        t = tp.append_column(self.key, self.column.get(tp))
        return (filters(t, self.filters) if self.filters else t)

class BooleanMaskNode(BaseNode):
    def __init__(self, parent, mask, cache_obj=None):
        self.parent, self.mask, self.cache_obj = parent, mask, cache_obj
        self.cache = (cache_obj != None)
        self.columns = parent.columns
        self.columns_source, self.columns_forward, self.filters_forward = parent.columns_source, parent.columns_forward, parent.filters_forward
    
    def fetch(self, verbose):
        tp = self.parent.get(verbose)
        t = tp.filter(self.mask.get(tp))
        return t

class FillNanNode(BaseNode):
    def __init__(self, parent, columns, value, cache_obj=None):
        self.parent, self.nan_columns, self.value, self.cache_obj = parent, columns, value, cache_obj
        self.cache = (cache_obj != None)
        self.check(needed=self.nan_columns, reference=parent.columns)
        self.columns = parent.columns

        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = parent.columns_source, list(set(parent.columns_forward + [c for c in self.nan_columns if c in parent.columns_source])), parent.filters_forward
    
    def fetch(self, verbose):
        t = self.parent.get(verbose)
        for c in self.nan_columns:
            arr = pa.compute.fill_null(t.column(c).combine_chunks(), pa.scalar(self.value))
            t = t.drop([c])
            t = t.append_column(c, arr)
        return t

class CastNode(BaseNode):
    def __init__(self, parent, dtypes, cache_obj=None):
        self.parent, self.dtypes, self.cache_obj = parent, dtypes, cache_obj
        self.cache = (cache_obj != None)
        self.check(needed=self.dtypes.keys(), reference=parent.columns)
        self.columns = parent.columns

        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = parent.columns_source, list(set(parent.columns_forward + [c for c in list(dtypes.keys()) if c in parent.columns_source])), parent.filters_forward
    
    def fetch(self, verbose):
        t = self.parent.get(verbose)
        for c, tp in self.dtypes.items():
            arr = pa.array(t.column(c).to_numpy().astype(tp))
            t = t.drop([c])
            t = t.append_column(c, arr)
        return t

class DropNode(BaseNode):
    def __init__(self, parent, columns, cache_obj=None):
        self.parent, self.drop_columns, self.cache_obj = parent, (columns if isinstance(columns, list) else [columns]), cache_obj
        self.cache = (cache_obj != None)
        self.check(needed=self.drop_columns, reference=parent.columns)
        self.columns = [c for c in parent.columns if c not in self.drop_columns]

        # Forward propagation of nodes
        self.columns_source, self.columns_forward, self.filters_forward = parent.columns_source, parent.columns_forward, parent.filters_forward
    
    def fetch(self, verbose):
        t = self.parent.get(verbose)
        return t.drop(self.drop_columns)


import pyarrow as pa
import numpy as np

# Numerical & logical operations 
class ColumnNode():
    def __init__(self, key, required, func=None, depth=0, boolean=False):
        self.key, self.required, self.depth, self.boolean = key, required, depth, boolean
        self.f = (lambda t: t[self.key] if not func else func)

    def get(self, t):
        f = self.f
        for _ in range(self.depth):
            f = self.f(t)
        return f(t)

    def __getitem__(self, key):
        keyn = self.key + '[' + str(key) + ']'
        if isinstance(key, str):
            f = lambda t: self.get(t).combine_chunks().field(key) # Reference to a Struct Column
        else:
            raise Exception("__getitem__ currently only defined for struct fields")
        return self.__class__(key=keyn, required=[r + '.' + key for r in self.required], func=f, depth=self.depth + 1)

    def breed(self, op, other_key, f, required=[], boolean=False):
        key = '(' + self.key + op + other_key + ')'
        return self.__class__(key=key, required=self.required + required, func=f, depth=self.depth + 1, boolean=boolean)

    # Numerical
    def __add__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.add(self.get(t), other.get(t))
            return self.breed('+', other.key, f, required=other.required)
        else:
            f = lambda t: pa.compute.add(self.get(t), pa.scalar(other))
            return self.breed('+', str(other), f)

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.subtract(self.get(t), other.get(t))
            return self.breed('-', other.key, f, required=other.required)
        else:
            f = lambda t: pa.compute.subtract(self.get(t), pa.scalar(other))
            return self.breed('-', str(other), f)

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.multiply(self.get(t), other.get(t))
            return self.breed('*', other.key, f, required=other.required)
        else:
            f = lambda t: pa.compute.multiply(self.get(t), pa.scalar(other))
            return self.breed('*', str(other), f)

    def __truediv__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.divide(self.get(t), other.get(t))
            return self.breed('/', other.key, f, required=other.required)
        else:
            f = lambda t: pa.compute.divide(self.get(t), pa.scalar(other))
            return self.breed('/', str(other), f)

    def __pow__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.array(np.power(self.get(t).to_numpy(), other.get(t).to_numpy()))
            return self.breed('**', other.key, f, required=other.required)
        else:
            f = lambda t: pa.array(np.power(self.get(t), other))
            return self.breed('**', str(other), f)
    
    # Comparison operations
    def __lt__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.less(self.get(t), other.get(t))
            return self.breed('<', other.key, f, required=other.required, boolean=True)
        else:
            f = lambda t: pa.compute.less(self.get(t), pa.scalar(other))
            return self.breed('<', str(other), f, boolean=True)

    def __le__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.less_equal(self.get(t), other.get(t))
            return self.breed('<', other.key, f, required=other.required, boolean=True)
        else:
            f = lambda t: pa.compute.less_equal(self.get(t), pa.scalar(other))
            return self.breed('<', str(other), f, boolean=True)

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.greater(self.get(t), other.get(t))
            return self.breed('>', other.key, f, required=other.required, boolean=True)
        else:
            f = lambda t: pa.compute.greater(self.get(t), pa.scalar(other))
            return self.breed('>', str(other), f, boolean=True)

    def __ge__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.greater_equal(self.get(t), other.get(t))
            return self.breed('>=', other.key, f, required=other.required, boolean=True)
        else:
            f = lambda t: pa.compute.greater_equal(self.get(t), pa.scalar(other))
            return self.breed('>=', str(other), f, boolean=True)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.equal(self.get(t), other.get(t))
            return self.breed('=', other.key, f, required=other.required, boolean=True)
        else:
            f = lambda t: pa.compute.equal(self.get(t), pa.scalar(other))
            return self.breed('=', str(other), f, boolean=True)

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.compute.not_equal(self.get(t), other.get(t))
            return self.breed('!=', other.key, f, required=other.required, boolean=True)
        else:
            f = lambda t: pa.compute.not_equal(self.get(t), pa.scalar(other))
            return self.breed('!=', str(other), f, boolean=True)

    # Logical operators
    def __invert__(self):
        f = lambda t: pa.compute.invert(self.get(t))
        return ColumnNode(key='~' + self.key, required=self.required, func=f, depth=self.depth + 1, boolean=True)

    def __and__(self, other):
        f = lambda t: pa.compute.and_(self.get(t), other.get(t))
        return self.breed('&', other.key, f, required=other.required, boolean=True)

    def __or__(self, other):
        f = lambda t: pa.compute.or_(self.get(t), other.get(t))
        return self.breed('|', other.key, f, required=other.required, boolean=True)

    def __xor__(self, other):
        f = lambda t: pa.compute.xor(self.get(t), other.get(t))
        return self.breed('^', other.key, f, required=other.required, boolean=True)

    # Rounding based
    def round(self, decimals=0):
        f = lambda t: pa.array(np.round(self.get(t).to_numpy(), decimals=decimals))
        return ColumnNode(key='round({}, {})'.format(self.key, decimals), required=self.required, func=f, depth=self.depth + 1)

    def ceil(self):
        f = lambda t: pa.array(np.ceil(self.get(t).to_numpy()))
        return ColumnNode(key='ceil({})'.format(self.key), required=self.required, func=f, depth=self.depth + 1)

    def floor(self):
        f = lambda t: pa.array(np.floor(self.get(t).to_numpy()))
        return ColumnNode(key='floor({})'.format(self.key), required=self.required, func=f, depth=self.depth + 1)

    # Casting
    def astype(self, dtype):
        f = lambda t: pa.array(self.get(t).to_numpy().astype(dtype))
        return ColumnNode(key='cast({} as {})'.format(self.key, dtype), required=self.required, func=f, depth=self.depth + 1)
    
    def cast(self, dtype):
        return self.astype(dtype)

    # SQL based
    def greatest(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.array(np.maximum(self.get(t).to_numpy(), other.get(t).to_numpy()))
            return ColumnNode(key='greatest({}, {})'.format(self.key, other.key), required=self.required + other.required, func=f, depth=self.depth + 1)
        else:
            f = lambda t: pa.array(np.maximum(self.get(t).to_numpy(), other))
            return ColumnNode(key='greatest({}, {})'.format(self.key, other), required=self.required, func=f, depth=self.depth + 1)

    def least(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.array(np.minimum(self.get(t).to_numpy(), other.get(t).to_numpy()))
            return ColumnNode(key='least({}, {})'.format(self.key, other.key), required=self.required + other.required, func=f, depth=self.depth + 1)
        else:
            f = lambda t: pa.array(np.minimum(self.get(t).to_numpy(), other))
            return ColumnNode(key='least({}, {})'.format(self.key, other), required=self.required, func=f, depth=self.depth + 1)

    def clip(self, a_min=None, a_max=None):
        f = lambda t: pa.array(np.clip(self.get(t).to_numpy(), a_min, a_max))
        return ColumnNode(key=self.key + '.clip({}, {})'.format(a_min, a_max), required=self.required, func=f, depth=self.depth + 1)

    def fillna(self, value):
        f = lambda t: pa.compute.fill_null(self.get(t), pa.scalar(other))
        return ColumnNode(key='{}.fillna({})'.format(self.key, other), required=self.required, func=f, depth=self.depth + 1)

    def coalesce(self, other):
        if isinstance(other, self.__class__):
            f = lambda t: pa.array(np.where(np.isnan(self.get(t)), other.get(t), self.get(t)))
            return ColumnNode(key='coalesce({}, {})'.format(self.key, other.key), required=self.required + other.required, func=f, depth=self.depth + 1)
        else:
            f = lambda t: pa.compute.fill_null(self.get(t), pa.scalar(other))
            return ColumnNode(key='coalesce({}, {})'.format(self.key, other), required=self.required, func=f, depth=self.depth + 1)

    # UDFs
    @classmethod
    def udf(cls, name, function, arguments):
        if isinstance(arguments, dict):
            f = lambda t: function(**{k: (v.get(t) if isinstance(v, ColumnNode) else v) for k,v in arguments.items()})
            keys = ', '.join([(v.key if isinstance(v, ColumnNode) else str(v)) for v in arguments.values()])
            depth = max([0] + [v.depth for v in arguments.values() if isinstance(v, ColumnNode)]) + 1
            required = list(set([r for v in arguments.values() if isinstance(v, ColumnNode) for r in v.required]))
        else:
            if isinstance(arguments, ColumnNode):
                f = lambda t: function(arguments.get(t))
                keys, depth, required = arguments.key, arguments.depth + 1, arguments.required
            else:
                f = lambda t: function(arguments)
                keys, depth, required = str(arguments), 0, []
        return cls(key=name + '(' + keys + ')', required=required, func=f, depth=depth)
